#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <memory.h>
#include <string.h>
#include <gtk/gtk.h>
#include <libgnomecanvas/libgnomecanvas.h>
#include <zlib.h>
#include <gdk/gdkx.h>
#include <X11/Xlib.h>

#include "xournal.h"
#include "xo-interface.h"
#include "xo-support.h"
#include "xo-callbacks.h"
#include "xo-misc.h"
#include "xo-file.h"

const char *tool_names[NUM_STROKE_TOOLS] = {"pen", "eraser", "highlighter"};
const char *color_names[COLOR_MAX] = {"black", "blue", "red", "green",
   "gray", "lightblue", "lightgreen", "magenta", "orange", "yellow", "white"};
const char *bgtype_names[3] = {"solid", "pixmap", "pdf"};
const char *bgcolor_names[COLOR_MAX] = {"", "blue", "pink", "green",
   "", "", "", "", "orange", "yellow", "white"};
const char *bgstyle_names[4] = {"plain", "lined", "ruled", "graph"};
const char *file_domain_names[3] = {"absolute", "attach", "clone"};

// creates a new empty journal

void new_journal(void)
{
  journal.npages = 1;
  journal.pages = g_list_append(NULL, new_page(&ui.default_page));
  journal.last_attach_no = 0;
  ui.pageno = 0;
  ui.layerno = 0;
  ui.cur_page = (struct Page *) journal.pages->data;
  ui.cur_layer = (struct Layer *) ui.cur_page->layers->data;
  ui.saved = TRUE;
  ui.filename = NULL;
  update_file_name(NULL);
}

// check attachment names

void chk_attach_names(void)
{
  GList *list;
  struct Background *bg;
  
  for (list = journal.pages; list!=NULL; list = list->next) {
    bg = ((struct Page *)list->data)->bg;
    if (bg->type == BG_SOLID || bg->file_domain != DOMAIN_ATTACH ||
        bg->filename->s != NULL) continue;
    bg->filename->s = g_strdup_printf("bg_%d.png", ++journal.last_attach_no);
  }
}

// saves the journal to a file: returns true on success, false on error

gboolean save_journal(const char *filename)
{
  gzFile f;
  struct Page *pg, *tmppg;
  struct Layer *layer;
  struct Item *item;
  int i, is_clone;
  char *tmpfn;
  GList *pagelist, *layerlist, *itemlist, *list;
  GtkWidget *dialog;
  
  f = gzopen(filename, "w");
  if (f==NULL) return FALSE;
  chk_attach_names();
  
  gzprintf(f, "<?xml version=\"1.0\" standalone=\"no\"?>\n"
     "<title>Xournal document - see http://math.mit.edu/~auroux/software/xournal/</title>\n"
     "<xournal version=\"" VERSION "\"/>\n");
  for (pagelist = journal.pages; pagelist!=NULL; pagelist = pagelist->next) {
    pg = (struct Page *)pagelist->data;
    gzprintf(f, "<page width=\"%.2f\" height=\"%.2f\">\n", pg->width, pg->height);
    gzprintf(f, "<background type=\"%s\" ", bgtype_names[pg->bg->type]); 
    if (pg->bg->type == BG_SOLID) {
      gzputs(f, "color=\"");
      if (pg->bg->color_no >= 0) gzputs(f, bgcolor_names[pg->bg->color_no]);
      else gzprintf(f, "#%08x", pg->bg->color_rgba);
      gzprintf(f, "\" style=\"%s\" ", bgstyle_names[pg->bg->ruling]);
    }
    else if (pg->bg->type == BG_PIXMAP) {
      is_clone = -1;
      for (list = journal.pages, i = 0; list!=pagelist; list = list->next, i++) {
        tmppg = (struct Page *)list->data;
        if (tmppg->bg->type == BG_PIXMAP && 
            tmppg->bg->pixbuf == pg->bg->pixbuf &&
            tmppg->bg->filename == pg->bg->filename)
          { is_clone = i; break; }
      }
      if (is_clone >= 0)
        gzprintf(f, "domain=\"clone\" filename=\"%d\" ", is_clone);
      else {
        if (pg->bg->file_domain == DOMAIN_ATTACH) {
          tmpfn = g_strdup_printf("%s.%s", filename, pg->bg->filename->s);
          if (!gdk_pixbuf_save(pg->bg->pixbuf, tmpfn, "png", NULL, NULL)) {
            dialog = gtk_message_dialog_new(GTK_WINDOW(winMain), GTK_DIALOG_MODAL,
              GTK_MESSAGE_ERROR, GTK_BUTTONS_OK, 
              "Could not write background '%s'. Continuing anyway.", tmpfn);
            gtk_dialog_run(GTK_DIALOG(dialog));
            gtk_widget_destroy(dialog);
          }
          g_free(tmpfn);
        }
        gzprintf(f, "domain=\"%s\" filename=\"%s\" ", 
          file_domain_names[pg->bg->file_domain], pg->bg->filename->s);
      }
    }
    gzprintf(f, "/>\n");
    for (layerlist = pg->layers; layerlist!=NULL; layerlist = layerlist->next) {
      layer = (struct Layer *)layerlist->data;
      gzprintf(f, "<layer>\n");
      for (itemlist = layer->items; itemlist!=NULL; itemlist = itemlist->next) {
        item = (struct Item *)itemlist->data;
        if (item->type == ITEM_STROKE) {
          gzprintf(f, "<stroke tool=\"%s\" color=\"", 
                          tool_names[item->brush.tool_type]);
          if (item->brush.color_no >= 0)
            gzputs(f, color_names[item->brush.color_no]);
          else
            gzprintf(f, "#%08x", item->brush.color_rgba);
          gzprintf(f, "\" width=\"%.2f\">\n", item->brush.thickness);
          for (i=0;i<2*item->path->num_points;i++)
            gzprintf(f, "%.2f ", item->path->coords[i]);
          gzprintf(f, "\n</stroke>\n");
        }
      }
      gzprintf(f, "</layer>\n");
    }
    gzprintf(f, "</page>\n");
  }
  gzclose(f);
  return TRUE;
}

// closes a journal: returns true on success, false on abort

gboolean close_journal(void)
{
  if (!ok_to_close()) return FALSE;
  
  // free everything...
  reset_selection();
  clear_redo_stack();
  clear_undo_stack();

  delete_journal(&journal);
  
  return TRUE;
  /* note: various members of ui and journal are now in invalid states,
     use new_journal() to reinitialize them */
}

// the XML parser functions for open_journal()

struct Journal tmpJournal;
struct Page *tmpPage;
struct Layer *tmpLayer;
struct Item *tmpItem;
char *tmpFilename;

GError *xoj_invalid(void)
{
  return g_error_new(G_MARKUP_ERROR, G_MARKUP_ERROR_INVALID_CONTENT, "Invalid file contents");
}

void xoj_parser_start_element(GMarkupParseContext *context,
   const gchar *element_name, const gchar **attribute_names, 
   const gchar **attribute_values, gpointer user_data, GError **error)
{
  int has_attr, i;
  char *ptr;
  struct Background *tmpbg;
  char *tmpbg_filename;
  GtkWidget *dialog;
  
  if (!strcmp(element_name, "title") || !strcmp(element_name, "xournal")) {
    if (tmpPage != NULL) {
      *error = xoj_invalid();
      return;
    }
    // nothing special to do
  }
  else if (!strcmp(element_name, "page")) { // start of a page
    if (tmpPage != NULL) {
      *error = xoj_invalid();
      return;
    }
    tmpPage = (struct Page *)g_malloc(sizeof(struct Page));
    tmpPage->layers = NULL;
    tmpPage->nlayers = 0;
    tmpPage->group = NULL;
    tmpPage->bg = g_new(struct Background, 1);
    tmpPage->bg->type = -1;
    tmpPage->bg->canvas_item = NULL;
    tmpPage->bg->pixbuf = NULL;
    tmpPage->bg->filename = NULL;
    tmpJournal.pages = g_list_append(tmpJournal.pages, tmpPage);
    tmpJournal.npages++;
    // scan for height and width attributes
    has_attr = 0;
    while (*attribute_names!=NULL) {
      if (!strcmp(*attribute_names, "width")) {
        if (has_attr & 1) *error = xoj_invalid();
        tmpPage->width = strtod(*attribute_values, &ptr);
        if (ptr == *attribute_values) *error = xoj_invalid();
        has_attr |= 1;
      }
      else if (!strcmp(*attribute_names, "height")) {
        if (has_attr & 2) *error = xoj_invalid();
        tmpPage->height = strtod(*attribute_values, &ptr);
        if (ptr == *attribute_values) *error = xoj_invalid();
        has_attr |= 2;
      }
      else *error = xoj_invalid();
      attribute_names++;
      attribute_values++;
    }
    if (has_attr!=3) *error = xoj_invalid();
  }
  else if (!strcmp(element_name, "background")) {
    if (tmpPage == NULL || tmpLayer !=NULL || tmpPage->bg->type >= 0) {
      *error = xoj_invalid();
      return;
    }
    has_attr = 0;
    while (*attribute_names!=NULL) {
      if (!strcmp(*attribute_names, "type")) {
        if (has_attr) *error = xoj_invalid();
        for (i=0; i<3; i++)
          if (!strcmp(*attribute_values, bgtype_names[i]))
            tmpPage->bg->type = i;
        if (tmpPage->bg->type < 0) *error = xoj_invalid();
        has_attr |= 1;
      }
      else if (!strcmp(*attribute_names, "color")) {
        if (tmpPage->bg->type != BG_SOLID) *error = xoj_invalid();
        if (has_attr & 2) *error = xoj_invalid();
        tmpPage->bg->color_no = COLOR_OTHER;
        for (i=0; i<COLOR_MAX; i++)
          if (!strcmp(*attribute_values, bgcolor_names[i])) {
            tmpPage->bg->color_no = i;
            tmpPage->bg->color_rgba = predef_bgcolors_rgba[i];
          }
        // there's also the case of hex (#rrggbbaa) colors
        if (tmpPage->bg->color_no == COLOR_OTHER && **attribute_values == '#') {
          tmpPage->bg->color_rgba = strtol(*attribute_values + 1, &ptr, 16);
          if (*ptr!=0) *error = xoj_invalid();
        }
        has_attr |= 2;
      }
      else if (!strcmp(*attribute_names, "style")) {
        if (tmpPage->bg->type != BG_SOLID) *error = xoj_invalid();
        if (has_attr & 4) *error = xoj_invalid();
        tmpPage->bg->ruling = -1;
        for (i=0; i<4; i++)
          if (!strcmp(*attribute_values, bgstyle_names[i]))
            tmpPage->bg->ruling = i;
        if (tmpPage->bg->ruling < 0) *error = xoj_invalid();
        has_attr |= 4;
      }
      else if (!strcmp(*attribute_names, "domain")) {
        if (tmpPage->bg->type <= BG_SOLID || (has_attr & 8))
          { *error = xoj_invalid(); return; }
        tmpPage->bg->file_domain = -1;
        for (i=0; i<3; i++)
          if (!strcmp(*attribute_values, file_domain_names[i]))
            tmpPage->bg->file_domain = i;
        if (tmpPage->bg->file_domain < 0)
          { *error = xoj_invalid(); return; }
        has_attr |= 8;
      }
      else if (!strcmp(*attribute_names, "filename")) {
        if (tmpPage->bg->type <= BG_SOLID || (has_attr != 9)) 
          { *error = xoj_invalid(); return; }
        if (tmpPage->bg->file_domain == DOMAIN_CLONE) {
          // filename is a page number
          i = strtol(*attribute_values, &ptr, 10);
          if (ptr == *attribute_values || i < 0 || i > tmpJournal.npages-2)
            { *error = xoj_invalid(); return; }
          tmpbg = ((struct Page *)g_list_nth_data(tmpJournal.pages, i))->bg;
          if (tmpbg->type != tmpPage->bg->type)
            { *error = xoj_invalid(); return; }
          tmpPage->bg->filename = refstring_ref(tmpbg->filename);
          tmpPage->bg->pixbuf = tmpbg->pixbuf;
          gdk_pixbuf_ref(tmpbg->pixbuf);
          tmpPage->bg->file_domain = tmpbg->file_domain;
        }
        else {
          tmpPage->bg->filename = new_refstring(*attribute_values);
          if (tmpPage->bg->file_domain == DOMAIN_ATTACH) {
            tmpbg_filename = g_strdup_printf("%s.%s", tmpFilename, *attribute_values);
            if (sscanf(*attribute_values, "bg_%d.png", &i) == 1)
              if (i > tmpJournal.last_attach_no) 
                tmpJournal.last_attach_no = i;
          }
          else tmpbg_filename = g_strdup(*attribute_values);
          tmpPage->bg->pixbuf = gdk_pixbuf_new_from_file(tmpbg_filename, NULL);
          if (tmpPage->bg->pixbuf == NULL) {
            dialog = gtk_message_dialog_new(GTK_WINDOW(winMain), GTK_DIALOG_MODAL,
              GTK_MESSAGE_WARNING, GTK_BUTTONS_OK, 
              "Could not open background '%s'. Setting background to white.",
              tmpbg_filename);
            gtk_dialog_run(GTK_DIALOG(dialog));
            gtk_widget_destroy(dialog);
            tmpPage->bg->pixbuf = gdk_pixbuf_new(GDK_COLORSPACE_RGB, FALSE, 8, 1, 1);
            gdk_pixbuf_fill(tmpPage->bg->pixbuf, 0xffffffff); // solid white
          }
          g_free(tmpbg_filename);
        }
        has_attr |= 16;
      }
      else *error = xoj_invalid();
      attribute_names++;
      attribute_values++;
    }
    if (tmpPage->bg->type < 0) *error = xoj_invalid();
    if (tmpPage->bg->type == BG_SOLID && has_attr != 7) *error = xoj_invalid();
    if (tmpPage->bg->type == BG_PIXMAP && has_attr != 25) *error = xoj_invalid();
  }
  else if (!strcmp(element_name, "layer")) { // start of a layer
    if (tmpPage == NULL || tmpLayer != NULL) {
      *error = xoj_invalid();
      return;
    }
    tmpLayer = (struct Layer *)g_malloc(sizeof(struct Layer));
    tmpLayer->items = NULL;
    tmpLayer->nitems = 0;
    tmpLayer->group = NULL;
    tmpPage->layers = g_list_append(tmpPage->layers, tmpLayer);
    tmpPage->nlayers++;
  }
  else if (!strcmp(element_name, "stroke")) { // start of a stroke
    if (tmpLayer == NULL || tmpItem != NULL) {
      *error = xoj_invalid();
      return;
    }
    tmpItem = (struct Item *)g_malloc(sizeof(struct Item));
    tmpItem->type = ITEM_STROKE;
    tmpItem->path = NULL;
    tmpItem->canvas_item = NULL;
    tmpLayer->items = g_list_append(tmpLayer->items, tmpItem);
    tmpLayer->nitems++;
    // scan for tool, color, and width attributes
    has_attr = 0;
    while (*attribute_names!=NULL) {
      if (!strcmp(*attribute_names, "width")) {
        if (has_attr & 1) *error = xoj_invalid();
        tmpItem->brush.thickness = strtod(*attribute_values, &ptr);
        if (ptr == *attribute_values) *error = xoj_invalid();
        has_attr |= 1;
      }
      else if (!strcmp(*attribute_names, "color")) {
        if (has_attr & 2) *error = xoj_invalid();
        tmpItem->brush.color_no = COLOR_OTHER;
        for (i=0; i<COLOR_MAX; i++)
          if (!strcmp(*attribute_values, color_names[i])) {
            tmpItem->brush.color_no = i;
            tmpItem->brush.color_rgba = predef_colors_rgba[i];
          }
        // there's also the case of hex (#rrggbbaa) colors
        if (tmpItem->brush.color_no == COLOR_OTHER && **attribute_values == '#') {
          tmpItem->brush.color_rgba = strtol(*attribute_values + 1, &ptr, 16);
          if (*ptr!=0) *error = xoj_invalid();
        }
        has_attr |= 2;
      }
      else if (!strcmp(*attribute_names, "tool")) {
        if (has_attr & 4) *error = xoj_invalid();
        tmpItem->brush.tool_type = -1;
        for (i=0; i<NUM_STROKE_TOOLS; i++)
          if (!strcmp(*attribute_values, tool_names[i])) {
            tmpItem->brush.tool_type = i;
          }
        if (tmpItem->brush.tool_type == -1) *error = xoj_invalid();
        has_attr |= 4;
      }
      else *error = xoj_invalid();
      attribute_names++;
      attribute_values++;
    }
    if (has_attr!=7) *error = xoj_invalid();
    // finish filling the brush info
    tmpItem->brush.thickness_no = 0;  // who cares ?
    tmpItem->brush.tool_options = 0;  // who cares ?
    if (tmpItem->brush.tool_type == TOOL_HIGHLIGHTER) {
      if (tmpItem->brush.color_no >= 0)
        tmpItem->brush.color_rgba &= HILITER_ALPHA_MASK;
    }
  }
}

void xoj_parser_end_element(GMarkupParseContext *context,
   const gchar *element_name, gpointer user_data, GError **error)
{
  if (!strcmp(element_name, "page")) {
    if (tmpPage == NULL || tmpLayer != NULL) {
      *error = xoj_invalid();
      return;
    }
    if (tmpPage->nlayers == 0 || tmpPage->bg->type < 0) *error = xoj_invalid();
    tmpPage = NULL;
  }
  if (!strcmp(element_name, "layer")) {
    if (tmpLayer == NULL || tmpItem != NULL) {
      *error = xoj_invalid();
      return;
    }
    tmpLayer = NULL;
  }
  if (!strcmp(element_name, "stroke")) {
    if (tmpItem == NULL) {
      *error = xoj_invalid();
      return;
    }
    update_item_bbox(tmpItem);
    tmpItem = NULL;
  }
}

void xoj_parser_text(GMarkupParseContext *context,
   const gchar *text, gsize text_len, gpointer user_data, GError **error)
{
  const gchar *element_name, *ptr;
  int n;
  
  element_name = g_markup_parse_context_get_element(context);
  if (element_name == NULL) return;
  if (!strcmp(element_name, "stroke")) {
    ptr = text;
    n = 0;
    while (text_len > 0) {
      realloc_cur_path(n/2 + 1);
      ui.cur_path.coords[n] = strtod(text, (char **)(&ptr));
      if (ptr == text) break;
      text_len -= (ptr - text);
      text = ptr;
      n++;
    }
    if (n<4 || n&1) { *error = xoj_invalid(); return; }
    tmpItem->path = gnome_canvas_points_new(n/2);
    g_memmove(tmpItem->path->coords, ui.cur_path.coords, n*sizeof(double));
  }
}

gboolean open_journal(char *filename)
{
  const GMarkupParser parser = { xoj_parser_start_element, 
                                 xoj_parser_end_element, 
                                 xoj_parser_text, NULL, NULL};
  GMarkupParseContext *context;
  GError *error;
  gboolean valid;
  gzFile f;
  char buffer[1000];
  int len;
  
  f = gzopen(filename, "r");
  if (f==NULL) return FALSE;
  
  context = g_markup_parse_context_new(&parser, 0, NULL, NULL);
  valid = TRUE;
  tmpJournal.npages = 0;
  tmpJournal.pages = NULL;
  tmpJournal.last_attach_no = 0;
  tmpPage = NULL;
  tmpLayer = NULL;
  tmpItem = NULL;
  tmpFilename = filename;
  error = NULL;

  while (valid && !gzeof(f)) {
    len = gzread(f, buffer, 1000);
    if (len<0) valid = FALSE;
    if (len<=0) break;
    valid = g_markup_parse_context_parse(context, buffer, len, &error);
  }
  gzclose(f);
  if (valid) valid = g_markup_parse_context_end_parse(context, &error);
  if (tmpJournal.npages == 0) valid = FALSE;
  g_markup_parse_context_free(context);
  if (!valid) {
    delete_journal(&tmpJournal);
    return FALSE;
  }
  
  ui.saved = TRUE; // force close_journal() to do its job
  close_journal();
  g_memmove(&journal, &tmpJournal, sizeof(struct Journal));
  ui.pageno = 0;
  ui.cur_page = (struct Page *)journal.pages->data;
  ui.layerno = ui.cur_page->nlayers-1;
  ui.cur_layer = (struct Layer *)(g_list_last(ui.cur_page->layers)->data);
  ui.saved = TRUE;
  update_file_name(filename);
  make_canvas_items();
  update_page_stuff();
  gtk_adjustment_set_value(gtk_layout_get_vadjustment(GTK_LAYOUT(canvas)), 0);
  return TRUE;
}

/************ file backgrounds *************/

struct Background *attempt_load_bg(char *filename)
{
  struct Background *bg;
  GdkPixbuf *pix;
  
  pix = gdk_pixbuf_new_from_file(filename, NULL);
  if (pix == NULL) return NULL;
  
  bg = g_new(struct Background, 1);
  bg->type = BG_PIXMAP;
  bg->canvas_item = NULL;
  bg->pixbuf = pix;
  bg->filename = new_refstring(filename);
  bg->file_domain = DOMAIN_ABSOLUTE;
  return bg;
}

struct Background *attempt_screenshot_bg(void)
{
  struct Background *bg;
  GdkPixbuf *pix;
  XEvent x_event;
  GError *error = NULL;
  GdkWindow *window;
  int x,y,w,h, status;
  unsigned int tmp;
  Window x_root, x_win;

  x_root = gdk_x11_get_default_root_xwindow();
  
  if (!XGrabButton(GDK_DISPLAY(), AnyButton, AnyModifier, x_root, 
      False, ButtonReleaseMask, GrabModeAsync, GrabModeSync, None, None))
    return NULL;

  XWindowEvent (GDK_DISPLAY(), x_root, ButtonReleaseMask, &x_event);
  XUngrabButton(GDK_DISPLAY(), AnyButton, AnyModifier, x_root);

  x_win = x_event.xbutton.subwindow;
  if (x_win == None) x_win = x_root;

  window = gdk_window_foreign_new_for_display(gdk_display_get_default(), x_win);
    
  gdk_window_get_geometry(window, &x, &y, &w, &h, NULL);
  
  pix = gdk_pixbuf_get_from_drawable(NULL, window,
    gdk_colormap_get_system(), 0, 0, 0, 0, w, h);
    
  if (pix == NULL) return NULL;
  
  bg = g_new(struct Background, 1);
  bg->type = BG_PIXMAP;
  bg->canvas_item = NULL;
  bg->pixbuf = pix;
  bg->filename = new_refstring(NULL);
  bg->file_domain = DOMAIN_ATTACH;
  return bg;
}

